"""MGZ database API."""

import logging
import multiprocessing
import json
import os
import sys
import tempfile
import zipfile

import iso8601
import requests_cache

import voobly
from mgzdb import queries, platforms
from mgzdb.add import add_file
from mgzdb.compress import decompress
from mgzdb.schema import get_session, reset, File, Match
from mgzdb.util import get_file


LOGGER = logging.getLogger(__name__)
QUERY_MATCH = 'match'
QUERY_FILE = 'file'
QUERY_SERIES = 'series'
QUERY_SUMMARY = 'summary'


class API: # pylint: disable=too-many-instance-attributes
    """MGZ Database API."""

    def __init__(self, db_path, store_path, consecutive=False, callback=None, **kwargs):
        """Initialize sessions."""

        self.session, _ = get_session(db_path)
        self.process_args = (store_path,)
        self.temp_dir = tempfile.TemporaryDirectory()
        self.store = None
        self.store_path = store_path
        self.db_path = db_path
        self.callback = callback
        self.debug = None
        self.consecutive = consecutive
        self.platforms = platforms.factory(
            voobly_key=kwargs.get('voobly_key'),
            voobly_username=kwargs.get('voobly_username'),
            voobly_password=kwargs.get('voobly_password')
        )
        self.voobly_key = kwargs.get('voobly_key')
        self.voobly_username = kwargs.get('voobly_username')
        self.voobly_password = kwargs.get('voobly_password')
        self.pool = None
        self.total = 0
        LOGGER.info("connected to database @ %s", db_path)

    def start(self):
        """Start child processes."""
        def init_worker(function):
            function.connections = {
                'session': get_session(self.db_path)[0],
                'aoe2map': requests_cache.CachedSession(backend='memory'),
                'platforms': platforms.factory(
                    voobly_key=self.voobly_key,
                    voobly_username=self.voobly_username,
                    voobly_password=self.voobly_password
                )
            }
        if self.consecutive:
            init_worker(add_file)
            self.debug = add_file
        else:
            self.pool = multiprocessing.Pool(
                initializer=init_worker,
                initargs=(add_file, )
            )

    def finished(self):
        """Wait for child processes to end."""
        try:
            if not self.consecutive:
                self.pool.close()
                self.pool.join()
        except KeyboardInterrupt:
            print('user requested exit')
            sys.exit()
        finally:
            self.temp_dir.cleanup()

    def add_file(self, *args, **kwargs):
        """Add file via process pool."""
        self.total += 1
        if self.consecutive:
            self.debug(*self.process_args, *args, **kwargs)
        else:
            if not self.pool:
                raise ValueError('call start() first')

            self.pool.apply_async(
                add_file,
                args=(*self.process_args, *args),
                kwds=kwargs,
                callback=self._file_added,
                error_callback=self._critical_error
            )

    def add_match(self, platform, url, single_pov=False):
        """Add a match via platform url."""
        if isinstance(url, str):
            match_id = url.split('/')[-1]
        else:
            match_id = url
        try:
            match = self.platforms[platform].get_match(match_id)
        except voobly.VooblyError as error:
            LOGGER.error("failed to get match: %s", error)
            return
        except ValueError:
            LOGGER.error("not an aoc match: %s", match_id)
            return
        players = match['players']
        chose = None
        if single_pov:
            for player in players:
                if player['url']:
                    chose = player
                    break
            if not chose:
                return
            players = [chose]
        for player in players:
            if not player['url']:
                continue
            try:
                filename = self.platforms[platform].download_rec(player['url'], self.temp_dir.name)
            except RuntimeError:
                LOGGER.error("could not download valid rec: %s", match_id)
                continue
            self.add_file(
                os.path.join(self.temp_dir.name, filename),
                url,
                platform_id=platform,
                platform_match_id=match_id,
                played=match['timestamp'],
                ladder=match.get('ladder'),
                user_data=match['players']
            )

    def add_series(self, zip_path, series=None, series_id=None):
        """Add a series via zip file."""
        with zipfile.ZipFile(zip_path) as series_zip:
            LOGGER.info("[%s] opened archive", os.path.basename(zip_path))
            series_zip.extractall(self.temp_dir.name)
            for filename in sorted(series_zip.namelist()):
                if filename.endswith('/'):
                    continue
                LOGGER.info("[%s] processing member file %s", os.path.basename(zip_path), filename)
                self.add_file(
                    os.path.join(self.temp_dir.name, filename),
                    os.path.basename(zip_path),
                    series,
                    series_id
                )
            LOGGER.info("[%s] finished", os.path.basename(zip_path))

    def add_db(self, remote):
        """Add records from another database."""
        for match in remote.session.query(Match).all():
            for mgz in match.files:
                filename, data = remote.get(mgz.id)
                path = os.path.join(self.temp_dir.name, filename)
                with open(path, 'wb') as handle:
                    handle.write(data)
                series = match.series.name if match.series else None
                series_id = match.series.series_id if match.series else None
                self.add_file(
                    path,
                    mgz.id,
                    series,
                    series_id=series_id,
                    platform_id=match.platform_id,
                    platform_match_id=match.platform_match_id,
                    played=match.played,
                    ladder=match.ladder.name if match.ladder else None,
                    user_data=[{
                        'id': mgz.owner.voobly_user_id,
                        'clan': mgz.owner.voobly_clan_id,
                        'color_id': mgz.owner.color_id,
                        'rate_before': mgz.owner.rate_before,
                        'rate_after': mgz.owner.rate_after
                    }]
                )

    def add_archive(self, path, single_pov=False):
        """Add records from archive."""
        archive_path = os.path.abspath(os.path.expanduser(path))
        for platform in os.listdir(archive_path):
            LOGGER.info("[%s] starting platform", platform)
            for subdir in os.listdir(os.path.join(archive_path, platform)):
                for match_id in os.listdir(os.path.join(archive_path, platform, subdir)):
                    match_path = os.path.join(archive_path, platform, subdir, match_id)
                    if not os.path.exists(os.path.join(match_path, 'metadata.json')):
                        continue
                    for zipped in os.listdir(match_path):
                        if zipped.endswith('.zip'):
                            with zipfile.ZipFile(os.path.join(match_path, zipped), 'r') as zip_ref:
                                zip_ref.extractall(os.path.join(self.temp_dir.name, match_id))
                    for mgz in os.listdir(os.path.join(self.temp_dir.name, match_id)):
                        if not mgz.endswith('.mgz'):
                            continue
                        mgz_path = os.path.join(self.temp_dir.name, match_id, mgz)
                        LOGGER.info('[%s][%s] adding match: %s', platform, match_id, mgz_path)
                        match = json.loads(open(os.path.join(match_path, 'metadata.json'), 'r').read())
                        self.add_file(
                            mgz_path,
                            mgz,
                            None, # series
                            platform_id=platform,
                            platform_match_id=int(match_id),
                            ladder=match.get('ladder'),
                            played=iso8601.parse_date(match['timestamp']),
                            user_data=match['players']
                        )
                        if single_pov:
                            break

    def remove(self, file_id=None, match_id=None):
        """Remove a file, match, or series."""
        if file_id:
            obj = self.session.query(File).get(file_id)
            if obj:
                if len(obj.match.files) == 1:
                    obj = obj.match.files[0].match
                self.session.delete(obj)
                self.session.commit()
                return
        elif match_id:
            obj = self.session.query(Match).get(match_id)
            if obj:
                for mgz in obj.files:
                    self.session.delete(mgz)
                self.session.commit()
                with self.session.no_autoflush:
                    for team in obj.teams:
                        self.session.delete(team)
                    for player in obj.players:
                        self.session.delete(player)
                self.session.commit()
                self.session.delete(obj)
                self.session.commit()
                return
        print('not found')

    def get(self, file_id):
        """Get a file from the store."""
        mgz_file = self.session.query(File).get(file_id)
        return mgz_file.original_filename, decompress(get_file(self.store_path, mgz_file.filename))

    def reset(self):
        """Reset database."""
        reset(self.db_path)
        LOGGER.info("reset database")

    def search(self):
        """Search database."""
        return self.session.query(Match).all()

    def query(self, query_type, **kwargs):
        """Query database."""
        if query_type == QUERY_MATCH:
            return queries.get_match(self.session, kwargs['match_id'])
        if query_type == QUERY_FILE:
            return queries.get_file(self.session, kwargs['file_id'])
        if query_type == QUERY_SERIES:
            return queries.get_series(self.session, kwargs['series_id'])
        if query_type == QUERY_SUMMARY:
            return queries.get_summary(self.session)
        raise ValueError('unsupported query type')

    def _file_added(self, success): # pylint: disable=unused-argument
        """Callback when file is addded."""
        if self.callback:
            self.callback()

    def _critical_error(self, val): # pylint: disable=unused-argument
        """Handle critical errors from child processes."""
        print(val, type(val))
        if self.callback:
            self.callback()



"""CLI for MGZ database."""
import argparse
import json
import logging
import os

import coloredlogs
import tqdm
from mgzdb.api import API
from mgzdb.util import parse_series_path


CMD_QUERY = 'query'
CMD_ADD = 'add'
CMD_REMOVE = 'remove'
CMD_GET = 'get'
CMD_RESET = 'reset'
CMD_BOOTSTRAP = 'bootstrap'
SUBCMD_FILE = 'file'
SUBCMD_MATCH = 'match'
SUBCMD_SERIES = 'series'
SUBCMD_SUMMARY = 'summary'
SUBCMD_ARCHIVE = 'archive'
SUBCMD_DB = 'db'
DEFAULT_DB = 'sqlite:///data.db'


def main(args): # pylint: disable=too-many-branches
    """Handle arguments."""

    add_callback = None
    if args.progress:
        progress = tqdm.tqdm(unit='mgz')
        coloredlogs.set_level('CRITICAL')
        add_callback = progress.update

    db_api = API(
        args.database, args.store_path,
        voobly_key=args.voobly_key,
        voobly_username=args.voobly_username,
        voobly_password=args.voobly_password,
        consecutive=args.consecutive,
        callback=add_callback
    )

    # Add
    if args.cmd == CMD_ADD:

        db_api.start()

        # File
        if args.subcmd == SUBCMD_FILE:
            for rec in args.rec_path:
                db_api.add_file(rec, args.source, None)
                if args.progress:
                    progress.total = db_api.total

        # Match
        elif args.subcmd == SUBCMD_MATCH:
            for url in args.url:
                db_api.add_match(args.platform, url)
                if args.progress:
                    progress.total = db_api.total

        # Series
        elif args.subcmd == SUBCMD_SERIES:
            for path in args.zip_path:
                series, series_id = parse_series_path(path)
                db_api.add_series(
                    path, series, series_id
                )
                if args.progress:
                    progress.total = db_api.total

        # Database
        elif args.subcmd == SUBCMD_DB:
            remote_api = API(args.remote_db_url, args.remote_store_path)
            db_api.add_db(remote_api)
            if args.progress:
                progress.total = db_api.total

        # Archive
        elif args.subcmd == SUBCMD_ARCHIVE:
            db_api.add_archive(args.archive_path)
            if args.progress:
                progress.total = db_api.total

        db_api.finished()
        if args.progress:
            progress.close()

    # Remove
    elif args.cmd == CMD_REMOVE:
        db_api.remove(file_id=args.file, match_id=args.match)

    # Query
    elif args.cmd == CMD_QUERY:
        print(json.dumps(db_api.query(args.subcmd, **vars(args)), indent=2))

    # Get
    elif args.cmd == CMD_GET:
        filename, data = db_api.get(args.file)
        output_filename = args.output_path or filename
        if os.path.exists(output_filename):
            print('file already exists:', output_filename)
            return
        with open(output_filename, 'wb') as handle:
            handle.write(data)
        print(output_filename)

    # Reset
    elif args.cmd == CMD_RESET:
        if input('reset database completely? [y/N] ') == 'y':
            db_api.reset()


def setup():
    """Setup CLI."""
    coloredlogs.install(
        level='INFO',
        fmt='%(asctime)s [%(process)d]%(name)s %(levelname)s %(message)s'
    )
    logging.getLogger('paramiko').setLevel(logging.WARN)
    logging.getLogger('voobly').setLevel(logging.WARN)
    logging.getLogger('sqlalchemy').setLevel(logging.INFO)

    parser = argparse.ArgumentParser()
    default_file_path = os.path.abspath('.')

    # Global options
    parser.add_argument('-d', '--database', default=os.environ.get('MGZ_DB', DEFAULT_DB))
    parser.add_argument('-sp', '--store-path', default=os.environ.get('MGZ_STORE_PATH', default_file_path))
    parser.add_argument('-vk', '--voobly-key', default=os.environ.get('VOOBLY_KEY', None))
    parser.add_argument('-vu', '--voobly-username', default=os.environ.get('VOOBLY_USERNAME', None))
    parser.add_argument('-vp', '--voobly-password', default=os.environ.get('VOOBLY_PASSWORD', None))
    parser.add_argument('-c', '--consecutive', action='store_true', default=False)
    parser.add_argument('-p', '--progress', action='store_true', default=False)

    # Commands
    subparsers = parser.add_subparsers(dest='cmd')
    subparsers.required = True

    # "query" command
    query = subparsers.add_parser(CMD_QUERY)

    # "query" subcommands
    query_subparsers = query.add_subparsers(dest='subcmd')
    query_subparsers.required = True

    # "query match"
    query_match = query_subparsers.add_parser(SUBCMD_MATCH)
    query_match.add_argument('match_id', type=int)

    # "query file"
    query_file = query_subparsers.add_parser(SUBCMD_FILE)
    query_file.add_argument('file_id', type=int)

    # "query series"
    query_series = query_subparsers.add_parser(SUBCMD_SERIES)
    query_series.add_argument('series_id', type=int)

    # "query summary"
    query_subparsers.add_parser(SUBCMD_SUMMARY)

    # "add" command
    add = subparsers.add_parser(CMD_ADD)

    # "add" subcommands
    add_subparsers = add.add_subparsers(dest='subcmd')
    add_subparsers.required = True

    # "add file"
    add_file = add_subparsers.add_parser(SUBCMD_FILE)
    add_file.add_argument('-s', '--source', default='cli')
    add_file.add_argument('--series')
    add_file.add_argument('--tournament')
    add_file.add_argument('rec_path', nargs='+')

    # "add match"
    add_match = add_subparsers.add_parser(SUBCMD_MATCH)
    add_match.add_argument('platform')
    add_match.add_argument('url', nargs='+')

    # "add series"
    add_series = add_subparsers.add_parser(SUBCMD_SERIES)
    add_series.add_argument('zip_path', nargs='+')

    # "add database"
    add_db = add_subparsers.add_parser(SUBCMD_DB)
    add_db.add_argument('remote_db_url')
    add_db.add_argument('remote_store_path')

    # "add archive"
    add_archive = add_subparsers.add_parser(SUBCMD_ARCHIVE)
    add_archive.add_argument('archive_path')

    # "remove" command
    remove = subparsers.add_parser(CMD_REMOVE)
    remove_group = remove.add_mutually_exclusive_group(required=True)
    remove_group.add_argument('-f', '--file')
    remove_group.add_argument('-m', '--match')

    # "get" command
    get = subparsers.add_parser(CMD_GET)
    get.add_argument('file')
    get.add_argument('-o', '--output-path')

    # "reset" command
    subparsers.add_parser(CMD_RESET)

    # "bootstrap" command
    subparsers.add_parser(CMD_BOOTSTRAP)

    args = parser.parse_args()
    main(args)


if __name__ == '__main__':
    setup()



"""MGZ database API."""

import hashlib
import io
import logging
import os
import sys
import time
from datetime import timedelta

import pkg_resources
import requests
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from sqlalchemy.exc import IntegrityError

import voobly
import mgz.const
import mgz.summary

from aocref.model import Series, Dataset, EventMap
from mgzdb.platforms import PLATFORM_VOOBLY, VOOBLY_PLATFORMS
from mgzdb.schema import (
    Match, SeriesMetadata, File, Ladder, Player,
    Team, User
)
from mgzdb.util import parse_filename_timestamp, save_file
from mgzdb.compress import compress


LOGGER = logging.getLogger(__name__)
LOG_ID_LENGTH = 8
COMPRESSED_EXT = '.mgc'


def add_file(
        store_path, rec_path, reference,
        series=None, series_id=None, platform_id=None,
        platform_match_id=None, played=None, ladder=None,
        user_data=None
    ):
    """Wrapper around AddFile class for parallelization."""
    obj = AddFile(
        add_file.connections, store_path # pylint: disable=no-member
    )
    try:
        return obj.add_file(
            rec_path, reference, series_name=series,
            series_id=series_id, platform_id=platform_id,
            platform_match_id=platform_match_id, played=played,
            ladder=ladder, user_data=user_data
        )
    except KeyboardInterrupt:
        sys.exit()


class AddFile:
    """Add file to MGZ Database."""

    def __init__(self, connections, store_path):
        """Initialize sessions."""
        self.store_path = store_path
        self.session = connections['session']
        self.platforms = connections['platforms']
        self.aoe2map = connections['aoe2map']

    def add_file( # pylint: disable=too-many-return-statements, too-many-branches
            self, rec_path, reference, series_name=None,
            series_id=None, platform_id=None, platform_match_id=None,
            played=None, ladder=None, user_data=None
    ):
        """Add a single mgz file."""
        start = time.time()
        if not os.path.isfile(rec_path):
            LOGGER.error("%s is not a file", rec_path)
            return False

        original_filename = os.path.basename(rec_path)

        with open(rec_path, 'rb') as handle:
            data = handle.read()

        try:
            handle = io.BytesIO(data)
            summary = mgz.summary.Summary(handle, len(data))
            # Hash against body only because header can vary based on compression
            file_hash = hashlib.sha1(data[summary.body_position:]).hexdigest()
            log_id = file_hash[:LOG_ID_LENGTH]
            LOGGER.info("[f:%s] add started", log_id)
        except RuntimeError:
            LOGGER.error("[f] invalid mgz file")
            return False

        if self.session.query(File).filter_by(hash=file_hash).count() > 0:
            LOGGER.warning("[f:%s] file already exists", log_id)
            return False

        try:
            summary.work()
        except RuntimeError:
            LOGGER.error("[f:%s] failed to parse mgz", log_id)
            return False

        encoding = summary.get_encoding()
        if encoding == 'unknown':
            LOGGER.warning("[f:%s] unknown file text encoding", log_id)
            return False

        match_hash_obj = summary.get_hash()
        if not match_hash_obj:
            LOGGER.error("f:%s] not enough data to calculate safe match hash", log_id)
            return False
        match_hash = match_hash_obj.hexdigest()

        try:
            where = (Match.hash == match_hash)
            if platform_match_id:
                where |= (Match.platform_match_id == platform_match_id)
            match = self.session.query(Match).filter(where).one()
            LOGGER.info("[f:%s] match already exists (%d); appending", log_id, match.id)
        except MultipleResultsFound:
            LOGGER.error("[f:%s] mismatched hash and match id: %s, %s", log_id, match_hash, platform_match_id)
            return False
        except NoResultFound:
            LOGGER.info("[f:%s] adding match", log_id)
            if not played:
                played = parse_filename_timestamp(original_filename)
            try:
                match = self._add_match(
                    summary, played, match_hash, user_data, series_name,
                    series_id, platform_id, platform_match_id, ladder
                )
                if not match:
                    return False
                self._update_match_users(platform_id, match.id, user_data)
            except IntegrityError:
                LOGGER.error("[f:%s] constraint violation: could not add match", log_id)
                return False

        compressed_filename, compressed_size = self._handle_file(file_hash, data)

        try:
            new_file = self._get_unique(
                File, ['hash'],
                filename=compressed_filename,
                original_filename=original_filename,
                hash=file_hash,
                size=summary.size,
                compressed_size=compressed_size,
                encoding=encoding,
                language=summary.get_language(),
                reference=reference,
                match=match,
                owner_number=summary.get_owner(),
                parser_version=pkg_resources.get_distribution('mgz').version
            )
            self.session.add(new_file)
            self.session.commit()
        except RuntimeError:
            LOGGER.error("[f:%s] unable to add file, likely hash collision", log_id)
            return False

        LOGGER.info("[f:%s] add finished in %.2f seconds, file id: %d, match id: %d",
                    log_id, time.time() - start, new_file.id, match.id)
        return True

    def _add_match( # pylint: disable=too-many-branches, too-many-return-statements
            self, summary, played, match_hash, user_data,
            series_name=None, series_id=None, platform_id=None,
            platform_match_id=None, ladder=None
    ):
        """Add a match."""
        try:
            postgame = summary.get_postgame()
            duration = summary.get_duration()
        except RuntimeError:
            LOGGER.error("[m] failed to get duration")
            return False

        log_id = match_hash[:LOG_ID_LENGTH]
        rated = False
        from_voobly, ladder_name, rated, _ = summary.get_ladder()
        if ladder:
            rated = True
        if not ladder:
            ladder = ladder_name
        if from_voobly and not platform_id:
            platform_id = PLATFORM_VOOBLY
        settings = summary.get_settings()
        try:
            map_data = summary.get_map()
        except ValueError:
            LOGGER.error("[m:%s] has an invalid map", log_id)
            return False
        completed = summary.get_completed()
        restored, _ = summary.get_restored()
        has_postgame = bool(postgame)
        major_version, minor_version = summary.get_version()
        try:
            dataset_data = summary.get_dataset()
        except ValueError:
            LOGGER.error("[m:%s] dataset inconclusive", log_id)
            return False
        teams = summary.get_teams()
        diplomacy = summary.get_diplomacy()
        players = list(summary.get_players())
        mirror = summary.get_mirror()
        if platform_match_id:
            log_id += ':{}'.format(platform_match_id)

        if restored:
            LOGGER.error("[m:%s] is restored game", log_id)
            return False

        if not completed:
            LOGGER.error("[m:%s] was not completed", log_id)
            return False

        if user_data and len(players) != len(user_data):
            LOGGER.error("[m:%s] has mismatched user data", log_id)
            return False

        if user_data:
            from_rec = {}
            for player in players:
                from_rec[player['color_id']] = player['name']
            from_user_data = {}
            for user in user_data:
                from_user_data[user['color_id']] = user['username']
            strike = 0
            for color_id, name in from_rec.items():
                if (
                        color_id not in from_user_data or
                        from_user_data[color_id].lower() not in name.lower() and
                        name.lower() not in from_user_data[color_id].lower()
                    ):
                    strike += 1
            if strike >= len(players) / 2:
                LOGGER.error("[m:%s] has mismatched user data (transposition)", log_id)
                return False

        if platform_id:
            try:
                if platform_id in VOOBLY_PLATFORMS:
                    ladder_id = voobly.lookup_ladder_id(ladder)
                else:
                    ladder_id = 1 # fix..
                ladder = self._get_unique(
                    Ladder, keys=['id', 'platform_id'], id=ladder_id,
                    platform_id=platform_id, name=ladder
                )
            except ValueError:
                ladder_id = None
        else:
            ladder_id = None

        try:
            dataset = self.session.query(Dataset).filter_by(id=dataset_data['id']).one()
        except NoResultFound:
            LOGGER.error("[m:%s] dataset not supported: userpatch id: %s (%s)",
                         log_id, dataset_data['id'], dataset_data['name'])
            return False

        series, tournament, event, event_map_id = self._handle_series(series_id, series_name, map_data, log_id)
        if series:
            played = series.played

        match = self._get_unique(
            Match, ['hash', 'platform_match_id'],
            platform_match_id=platform_match_id,
            platform_id=platform_id,
            played=played,
            hash=match_hash,
            series=series,
            tournament=tournament,
            event=event,
            version=major_version,
            minor_version=minor_version,
            dataset=dataset,
            dataset_version=dataset_data['version'],
            ladder_id=ladder_id,
            rated=rated,
            builtin_map_id=map_data['id'],
            event_map_id=event_map_id,
            map_size_id=map_data['dimension'],
            map_name=map_data['name'],
            rms_seed=map_data['seed'],
            rms_zr=map_data['zr'],
            rms_custom=map_data['custom'],
            guard_state=map_data['modes']['guard_state'],
            direct_placement=map_data['modes']['direct_placement'],
            effect_quantity=map_data['modes']['effect_quantity'],
            fixed_positions=map_data['modes']['fixed_positions'],
            duration=timedelta(milliseconds=duration),
            completed=completed,
            restored=restored,
            postgame=has_postgame,
            type_id=settings['type'][0],
            difficulty_id=settings['difficulty'][0],
            population_limit=settings['population_limit'],
            map_reveal_choice_id=settings['map_reveal_choice'][0],
            speed_id=settings['speed'][0],
            cheats=settings['cheats'],
            lock_teams=settings['lock_teams'],
            starting_resources_id=settings['starting_resources'][0],
            starting_age_id=settings['starting_age'][0],
            victory_condition_id=settings['victory_condition'][0],
            team_together=settings['team_together'],
            all_technologies=settings['all_technologies'],
            lock_speed=settings['lock_speed'],
            multiqueue=settings['multiqueue'],
            diplomacy_type=diplomacy['type'],
            team_size=diplomacy.get('team_size'),
            mirror=mirror
        )

        winning_team_id = None
        for data in players:
            team_id = None
            for i, team in enumerate(teams):
                if data['number'] in team:
                    team_id = i
            if data['winner']:
                winning_team_id = team_id
            feudal_time = data['achievements']['technology']['feudal_time']
            castle_time = data['achievements']['technology']['castle_time']
            imperial_time = data['achievements']['technology']['imperial_time']
            self._get_unique(
                Team,
                ['match', 'team_id'],
                winner=(team_id == winning_team_id),
                match=match,
                team_id=team_id
            )
            player = self._get_unique(
                Player,
                ['match_id', 'number'],
                civilization_id=int(data['civilization']),
                team_id=team_id,
                match_id=match.id,
                dataset=dataset,
                platform_id=platform_id,
                human=data['human'],
                name=data['name'],
                number=data['number'],
                color_id=data['color_id'],
                start_x=data['position'][0],
                start_y=data['position'][1],
                winner=data['winner'],
                mvp=data['mvp'],
                score=data['score'],
                rate_snapshot=data['rate_snapshot'],
                military_score=data['achievements']['military']['score'],
                units_killed=data['achievements']['military']['units_killed'],
                hit_points_killed=data['achievements']['military']['hit_points_killed'],
                units_lost=data['achievements']['military']['units_lost'],
                buildings_razed=data['achievements']['military']['buildings_razed'],
                hit_points_razed=data['achievements']['military']['hit_points_razed'],
                buildings_lost=data['achievements']['military']['buildings_lost'],
                units_converted=data['achievements']['military']['units_converted'],
                economy_score=data['achievements']['economy']['score'],
                food_collected=data['achievements']['economy']['food_collected'],
                wood_collected=data['achievements']['economy']['wood_collected'],
                stone_collected=data['achievements']['economy']['stone_collected'],
                gold_collected=data['achievements']['economy']['gold_collected'],
                tribute_sent=data['achievements']['economy']['tribute_sent'],
                tribute_received=data['achievements']['economy']['tribute_received'],
                trade_gold=data['achievements']['economy']['trade_gold'],
                relic_gold=data['achievements']['economy']['relic_gold'],
                technology_score=data['achievements']['technology']['score'],
                feudal_time=timedelta(seconds=feudal_time) if feudal_time else None,
                castle_time=timedelta(seconds=castle_time) if castle_time else None,
                imperial_time=timedelta(seconds=imperial_time) if imperial_time else None,
                explored_percent=data['achievements']['technology']['explored_percent'],
                research_count=data['achievements']['technology']['research_count'],
                research_percent=data['achievements']['technology']['research_percent'],
                society_score=data['achievements']['society']['score'],
                total_wonders=data['achievements']['society']['total_wonders'],
                total_castles=data['achievements']['society']['total_castles'],
                total_relics=data['achievements']['society']['total_relics'],
                villager_high=data['achievements']['society']['villager_high']
            )
            if match.platform_id == PLATFORM_VOOBLY and not user_data:
                self._guess_match_user(player, data['name'])

        match.winning_team_id = winning_team_id
        return match

    def _update_match_users(self, platform_id, match_id, user_data):
        """Update Voobly User info on Match."""
        if user_data:
            for user in user_data:
                try:
                    player = self.session.query(Player).filter_by(match_id=match_id, color_id=user['color_id']).one()
                except NoResultFound:
                    LOGGER.error("failed to find p%d to update platform user data",
                                 user['color_id'] + 1)
                    continue
                LOGGER.info("[m:%s] updating platform user data for p%d",
                            player.match.hash[:LOG_ID_LENGTH], user['color_id'] + 1)
                self._get_unique(User, ['id', 'platform_id'], id=str(user['id']), platform_id=platform_id)
                player.user_id = str(user['id'])
                player.user_name = user.get('username')
                player.platform_id = platform_id
                player.rate_before = user.get('rate_before')
                player.rate_after = user.get('rate_after')
                if not player.rate_snapshot:
                    player.rate_snapshot = user.get('rate_snapshot')

    def _guess_match_user(self, player, name):
        """Guess Voobly User from a player name."""
        try:
            user_id = str(self.platforms[PLATFORM_VOOBLY].find_user(name.lstrip('+')))
            self._get_unique(
                User, ['id', 'platform_id'],
                id=user_id,
                platform_id=PLATFORM_VOOBLY
            )
            player.user_id = user_id
        except (voobly.VooblyError, requests.exceptions.ConnectionError) as error:
            LOGGER.warning("failed to lookup Voobly user: %s", error)

    def _handle_series(self, series_id, series_name, map_data, log_id):
        """Handle series-related tasks."""
        if series_id and series_name:
            series = self.session.query(Series).get(series_id)
            if series:
                self._get_unique(
                    SeriesMetadata,
                    ['series_id'],
                    name=series_name,
                    series_id=series.id
                )
                tournament = series.tournament
                event = tournament.event
                event_map = self.session.query(EventMap) \
                    .filter(EventMap.event_id == event.id) \
                    .filter(EventMap.name == map_data['name']) \
                    .one_or_none()
                if event_map:
                    event_map_id = event_map.id
                else:
                    event_map_id = None
                    LOGGER.warning("[m:%s] event map for %s not found: %s", log_id, event.id, map_data['name'])
                return series, tournament, event, event_map_id
        try:
            event_map = self.session.query(EventMap) \
                .filter(EventMap.name == map_data['name']) \
                .one()
            event_map_id = event_map.id
            LOGGER.info("[m:%s] guessed event %s for map %s", log_id, event_map.event_id, map_data['name'])
        except (NoResultFound, MultipleResultsFound):
            event_map_id = None
        return None, None, None, event_map_id

    def _handle_file(self, file_hash, data):
        """Handle file: compress and store."""
        compressed_filename = '{}{}'.format(file_hash, COMPRESSED_EXT)
        compressed_data = compress(io.BytesIO(data))
        destination = save_file(compressed_data, self.store_path, compressed_filename)
        LOGGER.info("[f:%s] copied to %s", file_hash[:LOG_ID_LENGTH], destination)
        return compressed_filename, len(compressed_data)

    def _get_by_keys(self, table, keys, **kwargs):
        """Get object by unique keys."""
        return self.session.query(table).filter_by(**{k:kwargs[k] for k in keys}).one()

    def _get_unique(self, table, keys=None, **kwargs):
        """Get unique object either by query or creation."""
        if not keys:
            keys = ['name']
        if not any([kwargs[k] is not None for k in keys]):
            return None
        try:
            return self._get_by_keys(table, keys, **kwargs)
        except NoResultFound:
            self.session.begin_nested()
            try:
                obj = table(**kwargs)
                self.session.add(obj)
                self.session.commit()
                return obj
            except IntegrityError:
                self.session.rollback()
                try:
                    return self._get_by_keys(table, keys, **kwargs)
                except NoResultFound:
                    raise RuntimeError



"""Compress and Decompress MGZ files.

Excellent compression can be achieved by decompressing the
zlib-compressed header, re-compressing it with lzma, and
then compressing the body separately with lzma.

Reverse the process by decompressing the lzma-compressed
header, re-compressing it with zlib, and then decompressing
the lzma-compressed body.
"""

import logging
import lzma
import struct
import time
import zlib


PREFIX_SIZE = 8
LOGGER = logging.getLogger(__name__)
ZLIB_WBITS = -15
LZMA_DICT_SIZE = 64 * 1024 * 1024
LZMA_FILTERS = [
    {
        'id': lzma.FILTER_LZMA2,
        'preset': 7 | lzma.PRESET_EXTREME,
        'dict_size': LZMA_DICT_SIZE
    }
]


def compress(data):
    """Compress from file."""
    start = time.time()
    header_len, _ = struct.unpack('<II', data.read(PREFIX_SIZE))
    zlib_header = data.read(header_len - PREFIX_SIZE)
    header = zlib.decompress(zlib_header, wbits=ZLIB_WBITS)
    lzma_header = lzma.compress(header, filters=LZMA_FILTERS)

    body = data.read()
    lzma_body = lzma.compress(body, filters=LZMA_FILTERS)

    size = PREFIX_SIZE + len(zlib_header) + len(body)
    new_size = PREFIX_SIZE + len(lzma_header) + len(lzma_body)

    LOGGER.info("compressed input to %.1f%% of original size (%d->%d) in %.2f seconds",
                (new_size / size) * 100, size, new_size, time.time() - start)
    return struct.pack('<II', len(lzma_header) + PREFIX_SIZE, 0) + lzma_header + lzma_body


def decompress(data):
    """Decompress from file."""
    start = time.time()

    header_len, _ = struct.unpack('<II', data.read(PREFIX_SIZE))
    lzma_header = data.read(header_len - PREFIX_SIZE)
    header = lzma.decompress(lzma_header)
    zlib_header = zlib.compress(header)[2:]

    body = lzma.decompress(data.read())

    LOGGER.info("decompressed in %.2f seconds", time.time() - start)
    return struct.pack('<II', len(zlib_header) + PREFIX_SIZE, 0) + zlib_header + body
